/***************************************************/
/* Make Better Decisions with Optimization         */
/* David R. Duling                                 */
/* SAS Global Forum 2015                           */
/***************************************************/
options source notes ; 

/***************************************************/
/* SECTION 1                                       */
/* create data for credit ratings                  */
/***************************************************/
data appdata;
	do case = 1 to 10000;

   		/* add Condition variables for applications */  
		age = 18 + floor(ranuni(123) * 58);                         
		income = max(18000,floor(10000+(ranuni(123) * age * 1500))); 
		if ranuni(123) <= 2 * abs(age-45)/100 then mortdue = 0;      
		else mortdue = round(2.5 * ranuni(123) * income, 1);         

		/* add credit Ratings */
		ranvar = ranuni(123);
		if income >= 50000 then do;  
			if ranvar <= (0.5 - abs(age-45)/300)then rating = 1;
			else if ranvar <= 0.8  then rating = 2;
			else if ranvar <= 0.9  then rating = 3;
			else if ranvar <= 0.95 then rating = 4;
			else rating = 5;
		end;
		else do;
			if ranvar <= (0.3 - abs(age-45)/150) then rating = 1;
			else if ranvar <= 0.7 then rating = 2;
			else if ranvar <= 0.85 then rating = 3;
			else if ranvar <= 0.92 then rating = 4;
			else rating = 5;
		end;
	output;
	end;
	drop ranvar; 
run;

/***************************************************/
/* SECTION 2.   CREATE BUSINESS RULES              */
/***************************************************/
%macro rules ; 
	accept= 0 ; 
	if age > AGE1    and  income > INC1         and rating = 1  then do ; accept=1 ; end ; 
	if age > AGE1+5  and  income > INC1         and rating = 2  then do ; accept=1 ; end ; 
	if age > AGE1    and  income > INC1+10000   and rating = 2  then do ; accept=1 ; end ; 
	if age > AGE1+10 and  income > INC1+40000   and rating = 3  then do ; accept=1 ; end ; 	
%mend rules ; 

/* test rules  */
%let AGE1= 30 ; 
%let INC1= 40000;
title 'Distribution Before Optimization' ;
data scores ; set appdata ; 
	AGE1=&AGE1. ; 	
	INC1=&INC1. ; 
	%rules ; 
run ; 
proc freq data=scores ; 
	table accept * rating / norow nopercent nocol nocum out=before ; 
run ; 
proc sgplot data=scores ; 
	scatter x=age y=income / group=rating ;
run ; 


/***************************************************/
/* SECTION 3. SETUP the OPTIMIZATION FUNCTIONS     */
/***************************************************/
/* Count the number of rows in appdata to determine the array size. */
proc sql noprint;  select count(*) into :NN from appdata; quit;

options cmplib = work.brm ;
proc fcmp outlib=work.brm.opt;

   /* Define an acceptance function based on the business rules */
   function acceptance(AGE1, INC1, rating, age, income);
		%rules ; 
	    return(accept);
   endsub;

   /* Define the objective function */
   function objfunc(AGE1, INC1);
      array _age[1] /nosym;		rc=read_array('appdata', _age, "age");
      array _income[1] /nosym;	rc=read_array('appdata', _income, "income");
      array _rating[1] /nosym;	rc=read_array('appdata', _rating, "rating");
      ACCEPTS = 0;
      do k = 1 to &NN;
         if (acceptance(AGE1, INC1, _rating[k], _age[k], _income[k])) then ACCEPTS = ACCEPTS + 1;
      end;
      return (ACCEPTS);
   endsub;

   /* Define the function for the constraint on RATING ratio */
   function RATINGC(AGE1, INC1);
      array _age[1] /nosym;		rc=read_array('appdata', _age, "age");
      array _income[1] /nosym;	rc=read_array('appdata', _income, "income");
      array _rating[1] /nosym;	rc=read_array('appdata', _rating, "rating");
      numerator = 0;
      denominator = 0;
      do k = 1 to &NN;
         if acceptance(AGE1, INC1, _rating[k], _age[k], _income[k]) then do;
            if (_rating[k] > 1) then numerator = numerator + 1;
            denominator = denominator + 1;
         end; 
      end;
      return (numerator / denominator);
   endsub;

   /* Define the function for the constring on MORTDUE ratio */
   function MORTDUEC(AGE1, INC1); 
      array _age[1] /nosym;		rc=read_array('appdata', _age, "age");
      array _income[1] /nosym;	rc=read_array('appdata', _income, "income");
      array _rating[1] /nosym;	rc=read_array('appdata', _rating, "rating");
      array _mortdue[1] /nosym; rc=read_array('appdata', _mortdue, "mortdue");
      numerator = 0;
      denominator = 0;
      do k = 1 to &NN;
         if acceptance(AGE1, INC1, _rating[k], _age[k], _income[k]) then do;
            if (_mortdue[k] >= 100000) then numerator = numerator + 1;
            denominator = denominator + 1;
         end;
      end;
      return (numerator / denominator);
   endsub;
quit;

/***************************************************/
/* SECTION 4.  SETUP the OPTIMIZATION PARAMETERS   */
/***************************************************/

/* Objective data set (Maximize ACCEPTS) */
data objdata;
   input _id_ $ _function_ $ _sense_ $ ;
   datalines;
ACCEPTS objfunc max 
; run ; 

/* Variables dataset to indicate the variables and their bounds. 
	AGE1:  range for initial AGE threshold  
	INC1:  range for initial INCOME threshold 
*/
data vardata;
   input _id_ $ _lb_ _ub_ ;
   datalines;
AGE1 20 30
INC1 20000 40000
; run ;

/* Constraints data set 
	RATINGC: The ratio of cases accepted with RATING > 1 must not be exceed 50%. 
	MORTDUEC: the ratio of cases accepted with MORTDUE >= 100,000. nust not exceed 50% 
*/
data condata;
   input _id_ $ _lb_ _ub_;
   datalines;
RATINGC 0 0.50
MORTDUEC 0 0.50
run;

/***************************************************/
/* SECTION 4.  RUN THE OPTIMIZATION                */
/***************************************************/
title 'Optimize Business Rule Parameters' ;
proc optlso maxgen=100 lastgen=lastgen 
   variables = vardata
   objective = objdata 
   nlincon = condata
   primalout = solution
   feastol = 0.00001
   cacheout=cache
   ;
   readarray appdata;
   performance nthreads = 8; 
run;
%put &_OROPTLSO_ ; 

/***************************************************/
/* SECTION 5. DISPLAY RESULTS                      */
/***************************************************/

title1 'Optmization Results' ;
/* show results of the solution */
proc print data=solution label  noobs ; var _id_ _value_ ; run;

/* select optmial values from solution */
proc sql noprint ; 
	select _value_ into :AGE1 from solution where _id_ eq 'AGE1' ; 
	select _value_ into :INC1 from solution where _id_ eq 'INC1' ; 
	select _value_ into :RAT  from solution where _id_ eq 'RATINGC' ;
	select _value_ into :MORT from solution where _id_ eq 'MORTDUEC' ;
	select _value_ into :OBJ  from solution where _id_ eq '_obj_' ;
	select _value_ into :INF  from solution where _id_ eq '_inf_' ;
quit ;

/* score application data with optimal values */
data scores ; set appdata ; 
	AGE1= &AGE1. ; 	INC1= &INC1. ; 
	%rules ;
run ; 

/* generate plots on results */
proc sgplot data= scores ; 
	title2 "Optimal Age Threshold: &AGE1" ;
	histogram age / group= accept transparency=0.3 nbins=32 ;
	refline &AGE1. %sysevalf(&AGE1.+5) %sysevalf(&AGE1.+10) / axis=X lineattrs= (thickness=4 color=black)  ;
proc sgplot data= scores ; 
	title2 "Optimal Income Threshold: &INC1" ;
	histogram income / group= accept transparency=0.3 nbins=32 ;
	refline &INC1. %sysevalf(&INC1.+10000) %sysevalf(&INC1.+40000) / axis=X lineattrs= (thickness=4 color=black)  ;
proc sgplot data= scores ; 
	title2 "Rate of Accepts with Mortgages with value > $100,000 = &MORT." ;
	histogram mortdue / group= accept transparency=0.3 nbins=32 ;
	refline 100000 / axis=X lineattrs= (thickness=4 color=black)  ;
proc sgplot data= scores ; 
	title2 "Rate of Accepts with Ratings > 1 = &RAT." ;
	vbar rating / group= accept ; 
run ; 

/* create comparison between before and after optimization */
proc freq data=scores noprint ; 
	table accept * rating / norow nocol nopercent nocum out=after ; 
run ; 
data compare ; 
	merge before(rename=(count=before) drop= percent) 
			after(rename=(count=after) drop= percent) ;
	by accept rating ; 
	delta= after - before ;
run ; 
proc print data=compare noobs ; where accept eq 1 ; sum before after delta ; run ; 

/***************************************************/
/* BONUS:  OPTIMIZATION ANALYSIS                   */
/***************************************************/
proc transpose data=cache out=history ;
	by _sol_ ; 
	id _id_ ; 
	var _value_ ; 
run ; 
proc hpbin data=history quantile numbin=6 out=history_bins; 
	id _all_ ;
	input _obj_ _inf_ ; 
run ;
/***************************************************/
title 'Objective vs. Infeasibility' ;
proc sgplot data=history_bins ; 
	scatter x=_inf_  y=_obj_ / GROUP= bin__obj_ ; 
	refline &OBJ.  / axis=y ; 
run ;
title 'History for AGE1 parameter' ;
proc sgplot data=history_bins ; 
	scatter x=AGE1  y=_obj_  / GROUP= bin__obj_; 
	refline &AGE1. / axis=x ;
	refline &OBJ.  / axis=y ; 
run ;
title 'History for INC1 parameter' ;
proc sgplot data=history_bins ; 
	scatter x=INC1  y=_obj_ / GROUP= bin__obj_; 
	refline &INC1. / axis=x ;
	refline &OBJ.  / axis=y ; 
run ;
title 'AGE1 versus INC1 parameters' ;
proc sgplot data=history_bins ; 
	scatter x=AGE1 y=INC1 / GROUP= BIN__obj_ ;
	refline &AGE1.   / axis=x  lineattrs= (thickness=2 ) ; 
	refline &INC1.   / axis=y  lineattrs= (thickness=2 ) ; 
run ;
title 'Nonlinear constraints distribution' ;
proc sgplot data=history_bins ; 
	scatter x=MORTDUEC y=RATINGC / GROUP= bin__obj_ ; 
	refline &MORT. / axis=x  lineattrs= (thickness=2 ) ; 
	refline &RAT. / axis=y  lineattrs= (thickness=2 ) ; 
run ;







